@@ -3,6 +3,7 @@ import configparser
 import csv
 import json
 import pendulum
+import platform
 import os
 import shutil
 import stat
@@ -11,14 +12,17 @@ import yaml
 from ._validation_util import _validate_confirmation_inputs
 from .pretty_print import console_log, print_table, task_processing
 from datetime import datetime
-from grp import getgrgid
 from json.decoder import JSONDecodeError
 from pathlib import Path
-from pwd import getpwuid
-from typing import Any, NoReturn, Union
+from typing import Any, Dict, List, NoReturn, Union
 from yaml.parser import ParserError
 from yaml.scanner import ScannerError
 
+# the following libraries help resolve directory and file ownership on non-windows platforms
+if platform.system() != 'Windows':
+    from grp import getgrgid
+    from pwd import getpwuid
+
 
 class FileMuncher:
     """
@@ -44,6 +48,7 @@ class FileMuncher:
 
         self.base_dir = Path.cwd()   # current working directory
         self.home_dir = Path.home()  # user home folder
+        self.custom_path = {}        # placeholder to store custom paths
 
     def _convert_path_type(self, path_to_convert: Union['Path', str]) -> 'Path':
         """
@@ -88,55 +93,103 @@ class FileMuncher:
 
         # mapping of supported file types
         supported_file_extensions = {
-            'csv': ['csv'],                 # return csv if extension == csv
-            'ini': ['cfg', 'conf', 'ini'],  # return ini if extension == cfg | conf | ini
-            'json': ['json'],               # return json if extension == json
-            'txt': ['nfo', 'text', 'txt'],  # return txt if extension == nfo | text | txt
-            'yaml': ['yaml', 'yml']         # return yaml if extension == yaml | yml
+            'csv': ['csv'],                                                           # return csv if extension == csv
+            'ini': ['cfg', 'conf', 'ini'],                                            # return ini if extension == cfg | conf | ini
+            'json': ['json'],                                                         # return json if extension == json
+            'txt': ['nfo', 'text', 'txt'],                                            # return txt if extension == nfo | text | txt
+            'yaml': ['yaml', 'yml']                                                   # return yaml if extension == yaml | yml
         }
 
-        file_extension = path_to_file.suffix[1:]  # get the extension and remove the '.' from the name
-        if file_extension == '':
-            raise AttributeError(f'No extension provided for path {file_extension}')
+        file_extension = path_to_file.suffix[1:]                                      # get the extension and remove the '.' from the name
+        if file_extension == '':                                                      # if the path has no extension
+            raise AttributeError(f'No extension provided for path {file_extension}')  # raise the error
 
-        if verify_only:
-            return
+        if verify_only:                                                               # if only verifying
+            return                                                                    # return here with nothing
 
-        for std_type, alt_types in supported_file_extensions.items():  # begin check against alt_types
-            if file_extension in alt_types:                            # if the file_extension is one of the alt_types
-                return std_type                                        # return the normlaized std_type
+        for std_type, alt_types in supported_file_extensions.items():                 # begin check against alt_types
+            if file_extension in alt_types:                                           # if the file_extension is one of the alt_types
+                return std_type                                                       # return the normlaized std_typeß
 
-        # if no match is found -> raise error
-        raise TypeError(f"File extension '{file_extension}' is not supported.")
+        raise TypeError(f"File extension '{file_extension}' is not supported.")       # if no conditions are met then return error
 
     @staticmethod
-    def _read_csv_file(path_to_file: Union['Path', str]) -> Union[Any, NoReturn]:
+    def _get_path_owner(owner_id: int, id_type: str) -> Union[int, None, NoReturn, str]:
+        """
+        Convert user and group uid to a name. Only works for non-Windows platforms.
+
+        Args:
+            owner_id (str): the uid as reported by os.stat
+            id_type (str): the type of id to resolve. user or group
+
+        Returns:
+            owner_name (str): the translated owner id. if the platform is Windows then the uid will be returned back
+        """
+
+        if platform.system() == 'Windows':                               # if running on a Windows machine
+            return owner_id                                              # return back the id
+
+        else:                                                            # if anything other than Windows
+            owner_name = {
+                'user': lambda: getpwuid(owner_id).pw_name,              # user owner of the path
+                'group': lambda: getgrgid(owner_id).gr_name              # group owner of the path
+            }.get(id_type, lambda: None)()                               # check for the id_type or return None
+
+            if owner_name is None:                                       # if the owner_name not in the map
+                raise KeyError(f'{id_type} is not a valid owner type.')  # raise an error
+
+            return owner_name                                            # return the translated owner name
+
+    @staticmethod
+    def _read_csv_file(path_to_file: Union['Path', str], options: Dict[str, Any] = dict()) -> Union[List[Union[dict[str, str], List[str]]], NoReturn]:
         """
         Read in the contents of a .csv file and return as a list of dictionaries.
 
         Args:
             path_to_file (Path | str): path to json file
 
+        Optional Args:
+            options (dict): options to control the way a csv file is read; default: no options set
+                Available Options:
+                    no_headers: set to True if the csv does not have headers
+
         Returns:
-            contents (dict): json contents read in as a list of dictionaries
+            contents (list): default is list of dictionaries
+                (list of dicts) returns list of dictionaries if no_headers option is not set or is False
+                (list of lists) return list of lists if no_headers option is set to True
+
         """
 
-        contents = []                                         # place holder for all csv contents
-        index = int(0)                                        # index used to validate headers only one time
-        with open(path_to_file, encoding='utf-8') as infile:  # open the file
-            csvReader = csv.DictReader(infile)                # read the csv file
+        if not options:                                                    # if no options specified then create default
+            options = {
+                'no_headers': False                                        # require headers by default
+            }
 
-            for rows in csvReader:                            # begin loop over all csv rows
+        index = int(0)                                                     # index used to validate headers only one time
+        contents = []                                                      # place holder for all csv contents
+        with open(path_to_file, encoding='utf-8') as infile:               # open the file
+            csvReader = csv.DictReader(infile)                             # read the csv file
+
+            for rows in csvReader:                                         # begin loop over all csv rows
                 if index == 0:
-                    if None in rows.keys():                   # check that all columns have a header
+                    index += 1                                             # increment the index so we dont check again
+                    if None in rows.keys() and not options['no_headers']:  # check that all columns have a header
                         # if a column has no header then raise the error
                         raise IndexError(f'Column header not defined for all columns. File: {path_to_file}')
-                    index += 1                                # increment the index so we dont check again
 
-                contents.append(rows)                         # add the row to content
-        infile.close()                                        # close the file
+                    elif options['no_headers']:                              # if no_headers is True
+                        contents.append([value for value in rows.keys()])    # rows.keys is equal to the values from row one
+                        contents.append([value for value in rows.values()])  # rows.values is equal to the values from the second row
+                        continue                                             # skip to row 3
 
-        return contents                                       # return
+                if options['no_headers']:                                  # if no_headers is True
+                    contents.append([value for value in rows.values()])    # append the values only
+
+                else:
+                    contents.append(rows)                                  # add the row to content
+        infile.close()                                                     # close the file
+
+        return contents                                                    # return
 
     @staticmethod
     def _read_ini_file(path_to_file: Union['Path', str]) -> dict[str, dict[str, str]]:
@@ -433,6 +486,41 @@ class FileMuncher:
         self._get_file_extension(file_to_create, verify_only=True)  # verify the file path has an extension
         file_to_create.touch(exist_ok=True)                         # create the file
 
+    def get_directory_contents(self, path_to_directory: Union['Path', str], recursive: bool = False) -> Dict[str, Any]:
+        """
+        List the contents of a single directory or all directories recursively.
+
+        Args:
+            path_to_directory (Path | str): path to the directory
+            recursive (bool): set to True to recursively list all contents of each directory in and under the given path; defaults to False
+
+        Returns:
+            directory_contents (dict): dictionary of the folders and files
+        """
+
+        path_to_directory = self._convert_path_type(path_to_directory)                  # normalize the path
+
+        contents = {                                                                    # create the contents container
+            path_to_directory.name: {                                                   # use the path name as parent key
+                'directories': [] if recursive is False else {},                        # if recursive is False the list directories else create a dict to gather the contents
+                'files': []                                                             # store file names in a list
+            }
+        }
+
+        if self._verify_path_exists(path_to_directory) and path_to_directory.is_dir():  # if the path exists and the path is a directory
+            for item in path_to_directory.iterdir():                                    # check each item in the directory
+                sub_name = 'files'                                                      # default sub_name to files
+                if item.is_dir():                                                       # is the current item a directory?
+                    sub_name = 'directories'                                            # if yes then change sub_name to directory
+
+                    if recursive:                                                       # if the item is a directory and recursive is True
+                        contents[path_to_directory.name][sub_name] = self.get_directory_contents(item, recursive)  # get the contents of the current directory
+                        continue
+
+                contents[path_to_directory.name][sub_name].append(item.name)            # add the file name to the list
+
+        return contents                                                                 # return the contents
+
     def get_path_stats(self, filepath: Union['Path', str], timezone: str = 'UTC') -> dict[str, str]:
         """
         Get additional file system details about a directory or file.
@@ -469,8 +557,8 @@ class FileMuncher:
         file_stats = {
             'type': 'file' if converted_path.is_file() else 'directory',                                  # file or directory
             'st_mode': stat.filemode(stats.st_mode),                                                      # read|write|execute permissions
-            'st_uid': getpwuid(stats.st_uid).pw_name,                                                     # user owner of the path
-            'st_gid': getgrgid(stats.st_gid).gr_name,                                                     # group owner of the path
+            'st_uid': self._get_path_owner(stats.st_uid, 'user'),                                         # user owner of the path
+            'st_gid': self._get_path_owner(stats.st_gid, 'group'),                                        # group owner of the path
             'st_size': stats.st_size,                                                                     # size in bytes
             'st_atime': pendulum.from_timestamp(stats.st_atime, timezone).format('YYYY-MM-DD HH:MM:SS'),  # time of most recent access in seconds
             'st_mtime': pendulum.from_timestamp(stats.st_atime, timezone).format('YYYY-MM-DD HH:MM:SS'),  # time of most recent content modification in seconds
@@ -479,7 +567,7 @@ class FileMuncher:
 
         return file_stats                                                                                 # return
 
-    def read_file(self, file_to_read: Union['Path', str]) -> Any:
+    def read_file(self, file_to_read: Union['Path', str], options: Dict[str, Any] = {}) -> Any:
         """
         Read in a file and get the contents. Supports the following file types:
 
@@ -494,6 +582,7 @@ class FileMuncher:
 
         Returns:
             * csv: list of dicts
+                * no_headers (option): list of lists
             * cfg, conf, ini: dict
             * json: dict
             * nfo, text, txt: list
@@ -512,7 +601,7 @@ class FileMuncher:
                     file_extension = str(self._get_file_extension(file_to_read))  # get the extension of the file
 
                     contents = {                                                  # -- extension to function map --
-                        'csv': lambda: self._read_csv_file(file_to_read),             # read csv contents
+                        'csv': lambda: self._read_csv_file(file_to_read, options),    # read csv contents
                         'ini': lambda: self._read_ini_file(file_to_read),             # read ini contents
                         'json': lambda: self._read_json_file(file_to_read),           # read json contents
                         'txt': lambda: self._read_txt_file(file_to_read),             # read txt contents
@@ -649,7 +738,11 @@ class FileMuncher:
         """
 
         converted_path = self._convert_path_type(attribute_path)                                   # normalize from str to Path
-        setattr(self, attribute_name, converted_path)                                              # create new Files attribute to provided attribute_path
+        if getattr(self, attribute_name, None) is not None:
+            setattr(self, attribute_name, converted_path)                                          # create new Files attribute to provided attribute_path
+
+        else:
+            self.custom_path[attribute_name] = converted_path
 
         if is_dir and is_file:                                                                     # if both is_dir and is_file flags are True
             raise TypeError('Both is_dir and is_file are true. Only one option may be selected.')  # raise error
